(use-modules (ellinika elmorph)
             (ellinika i18n)
	     (ellinika cgi)
	     (xmltools dict)
	     (gamma sql))

(use-syntax (ice-9 syncase))

; FIXME:
(ellinika-cgi-init dict-template-file-name)

(define (mk-dict-connect)
  (let ((db-connection #f))
    (lambda (. rest)
      (cond
       ((null? rest)
	(if (not db-connection)
	    (begin
	      (set! db-connection
		    (sql-open-connection
		     ellinika-sql-connection))
	      (sql-query db-connection "SET NAMES utf8")
	      )))
       (else
	(if db-connection
	    (sql-close-connection db-connection))
	(set! db-connection #f)))
      db-connection)))

(define dict-connect (mk-dict-connect))

(define (q-my-sql-query conn query)
  (catch #t
	 (lambda ()
	   (sql-query conn query))
	 (lambda args
	   '())))

(define (my-sql-query conn query)
;  (display "Q:")(display query)(newline)
  (sql-query conn query))


(define (sql-error-handler key func fmt fmtargs data)
  (format #t "<h1 class=\"error\">~A</h1>\n"
	  (_ "ΣΦΆΛΜΑ: σύνδεση με το λέξικο απέτυχε."))
  (apply format (current-error-port) fmt fmtargs))

(define (guess-verb-info verb)
  (let ((elverb (string->elstr verb)))
  (cond
   (else
    (let ((root (elstr-trim elverb -1)))
      (list "A"
	    #f
	    "ε"
	    #f))))))

(define (get-verb-info verb)
  (let ((conn (dict-connect)))
    (let ((vdb (my-sql-query
		conn
		(string-append
		 "SELECT conj,accmap,augment FROM verb \
WHERE verb='" verb "'"))))
      (if (null? vdb)
	  (guess-verb-info verb)
	  (let ((x (car vdb)))
	    (list
	     (list-ref x 0)
	     (list-ref x 1)
	     (or (list-ref x 2) "ε")
	     #f))))))

(define (complement-verb-info vinfo verb voice thema)
;  (format #t "COMPLEMENT ~S~%" thema)
  (let ((result (my-sql-query
		 (dict-connect)
		 (string-append
		 "SELECT root FROM irregular_root \
WHERE verb='" verb "' AND voice='" voice "' AND thema='" thema "'"))))
    (list-set! vinfo 3 
	       (if (not (null? result))
		   (caar result)
		   (let ((root (elstr-trim (string->elstr verb) -1)))
		     (cond
		      ((string=? thema "pres")
		       root)
		      ((or (string=? thema "aor") (string=? thema "sub"))
		       (elstr-thema-aoristoy root))
		      (else
		       #f)))))))

(define-syntax verb-info
  (syntax-rules ()
    ((verb-info #:conj v)
     (list-ref v 0))
    ((verb-info #:accmap v)
     (list-ref v 1))
    ((verb-info #:augment v)
     (list-ref v 2))
    ((verb-info #:root v)
     (list-ref v 3))))

(define-syntax conj-info
  (syntax-rules ()
    ((conj-info #:thema v)
     (list-ref v 0))
    ((conj-info #:suffix v)
     (list-ref v 1))
    ((conj-info #:accmap v)
     (list-ref v 2))
    ((conj-info #:particle v)
     (list-ref v 3))
    ((conj-info #:aux v)
     (list-ref v 4))
    ((conj-info #:auxtense v)
     (list-ref v 5))
    ((conj-info #:flect v)
     (list-tail v 6))
    ((conj-info #:sing 1 v)
     (list-ref v 7))
    ((conj-info #:sing 2 v)
     (list-ref v 8))
    ((conj-info #:sing 3 v)
     (list-ref v 9))
    ((conj-info #:plur 1 v)
     (list-ref v 10))
    ((conj-info #:plur 1 v)
     (list-ref v 11))
    ((conj-info #:plur 1 v)
     (list-ref v 12))))
    
(define (get-conj-info conj voice mode tense)
  (let ((conn (dict-connect)))
    (let ((answer (my-sql-query
		   conn
		   (string-append
      "SELECT c.thema,c.suffix,c.accmap,c.particle,c.aux,c.auxtense,\
f.sing1,f.sing2,f.sing3,f.plur1,f.plur2,f.plur3 \
FROM conjugation c, verbflect f \
WHERE c.conj='" conj "' AND c.voice='" voice "' AND c.mode='" mode
"' AND c.tense='" tense "' AND c.flect = f.ident"))))
      (if (null? answer)
	  #f
	  (car answer)))))

(define (force-string str)
  (if (elstr? str)
      (elstr->string str)
      str))

(define (force-elstr str)
  (if (string? str)
      (string->elstr str)
      str))

(define (apply-flect conj vinfo)
  (let ((root (verb-info #:root vinfo))
	(suffix (or (conj-info #:suffix conj) ""))
	(accmap (string->list (or (verb-info #:accmap vinfo)
				  (conj-info #:accmap conj)
				  "000000")))
	(augment ""))
;    (format #t "ROOT ~S, ACCMAP ~S~%" root accmap)
    (cond
     ((> (length accmap) 6)
      (set! accmap (list-head accmap 6))
      (set! augment (verb-info #:augment vinfo))))
    (let ((forms
	   (map
	    (lambda (flect acc)
	      (cond
	       ((char=? acc #\0)
		(let* ((rs (force-elstr root))
		       (suf (elstr-deaccent (elstr-append suffix flect)))
		       (result (elstr-append rs suf)))
		  (cond
		   ((or (= (elstr-accented-syllable rs) 0)
			(> (elstr-number-of-syllables suf) 2))
		    (let ((nsyl (elstr-number-of-syllables suf)))
		      (cond
		       ((= nsyl 1)
			result)
		       ((= nsyl 3)
			(elstr-set-accent result 3))
		       (else
			(elstr-set-accent result 2)))))
		   (else
		    result))))
	       ((char=? acc #\f)
		(elstr-append
		 (elstr-deaccent (elstr-append root suffix))
		 flect))
	       ((char=? acc #\s)
		(elstr-append
		 (elstr-deaccent (force-elstr root))
		 suffix
		 (elstr-deaccent (force-elstr flect))))
	       ((char=? acc #\-)
		#f)
	       ((char-numeric? acc)
		(let ((num (- (char->integer acc) (char->integer #\0))))
		  (let ((obj (elstr-append root suffix flect)))
		    (if (and augment (= (+ (elstr-number-of-syllables obj) 1)
					num))
			(set! obj (elstr-append augment obj)))
		    (elstr-set-accent! obj num)
		    obj)))
	       (else
		(error "invalid accent character" acc))))
	    (conj-info #:flect conj)
	    accmap)))
      (if (conj-info #:particle conj)
	  (map
	   (lambda (w)
	     (if w
		 (string-append
		  (conj-info #:particle conj) " " (force-string w))))
	   forms)
	  forms))))

(define (individual-verb verb voice mode tense)
  (let ((res (my-sql-query
	      (dict-connect)
	      (string-append
	       "SELECT f.sing1,f.sing2,f.sing3,f.plur1,f.plur2,f.plur3 \
FROM individual_verb i,verbflect f \
WHERE i.verb='" verb "' AND i.voice='" voice "' AND i.mode='" mode
"' AND i.tense = '" tense "' AND i.ident=f.ident"))))
    (if (not (null? res))
	(car res)
	#f)))

(define (conjugate verb voice mode tense . rest)
  (cond
   ((individual-verb verb voice mode tense) =>
    (lambda (res)
      res))
   (else
    (let* ((vinfo (get-verb-info verb))
	   (conj (get-conj-info (verb-info #:conj vinfo) voice mode tense)))
      (if (not conj)
	  (error "cannot obtain conjugation information for "
		 (verb-info #:conj vinfo) voice mode tense))
      (if (member #:nopart rest)
	  (list-set! conj 3 #f))
      (cond
       ((string=? (conj-info #:thema conj) "synt")
	(let ((form (list-ref
		     (conjugate verb "act" "sub" "Αόριστος" #:nopart) 2))
	      (part (conj-info #:particle conj)))
	  (map
	   (lambda (aux)
	     (elstr->string
	      (if part
		  (elstr-append part " " aux " " form)
		  (elstr-append aux " " form))))
	   (conjugate (conj-info #:aux conj) "act" "ind"
		      (conj-info #:auxtense conj))) ))
       (else
	;      (format #t "CONJ ~S~%" conj)
	(complement-verb-info vinfo verb voice (conj-info #:thema conj))
	(apply-flect conj vinfo)))))))

;;
;(display (verb-info "βρίσκω"))
;(newline)
;(display (verb-info "ανοίγω"))
;(newline)

(define (test-conjugation verb voice mode tense)
  (format #t "~A ~A ~A ~A: " verb voice mode tense)
  (display (map force-string (conjugate verb voice mode tense)))
  (newline)
  (gc))

(test-conjugation "είμαι" "act" "ind" "Ενεστώτας")
(test-conjugation "είμαι" "act" "ind" "Παρατατίκος")
(test-conjugation "είμαι" "act" "ind" "Μέλλοντας διαρκείας")
(test-conjugation "έχω" "act" "ind" "Ενεστώτας")
(test-conjugation "ανοίγω" "act" "ind" "Ενεστώτας")
(test-conjugation "ανοίγω" "act" "ind" "Αόριστος")
(test-conjugation "δένω" "act" "ind" "Αόριστος")
(test-conjugation "θέλω" "act" "ind" "Αόριστος")
(test-conjugation "θέλω" "act" "ind" "Παρατατικός")
(test-conjugation "βρίσκω" "act" "ind" "Ενεστώτας")
(test-conjugation "βρίσκω" "act" "ind" "Αόριστος")
(test-conjugation "βρίσκω" "act" "ind" "Μέλλοντας στιγμιαίος")
(test-conjugation "βρίσκω" "pas" "ind" "Αόριστος")
(test-conjugation "βρίσκω" "act" "sub" "Ενεστώτας")
(test-conjugation "βρίσκω" "act" "sub" "Αόριστος")
(test-conjugation "βρίσκω" "act" "ind" "Παρακείμενος")
(test-conjugation "βρίσκω" "act" "sub" "Παρακείμενος")

(newline)
